'--------------------------------------------------------------
' CP0046 - Quad Batteries Supervisor Firmware
' File: CP0046.BAS
' By Philippe Demerliac
' (c) 2012-2013 Cyrob
'------------------------------------------------------------------------------
' Version history
' Version Date          Author      Comment
'   1.0d1 01 Jan 2013   Phildem     Preliminary tests
'   1.0d2 03 Jan 2013   Phildem     Cal Mode And Battery Detection
'   1.0   06 Jan 2013   Phildem     First Release OK
'------------------------------------------------------------------------------
' Hardware connexions

' P0 Use all pins as inputs................................................
' All p0 pins are connected to ADC0804 Data Bus

' P1 Use ..................................................................

' Out DispDIn     P1.0  Activ Hi    MAX7219 DataIn
' Out DispClock   P1.1  Activ Hi    MAX7219 Clock
' Out DispLoad    P1.2  Activ Low   MAX7219 Load
' Out AdcRead     P1.3  Activ Low   ADC0804 Read & CS
' Out AdcWrite    P1.4  Activ Low   ADC0804 Write & CS
' In  AdcEOC      P1.5  Activ Low   ADC0804 INTR (End Of Conversion)
' In  DebugMode   P1.6  Activ Low   Indicate DBug Mode
' Out WdTrig      P1.7  Toggle      Max813L WDI Must be toggled every Sec

' P2 Use all pins as outputs ..............................................

' ActBat1   P2.0  A logic Low select the Battery n1
' ActBat2   P2.1  A logic Low select the Battery n2
' ActBat3   P2.2  A logic Low select the Battery n3
' ActBat4   P2.3  A logic Low select the Battery n4
' LedLow1   P2.4  A logic Hi light the Led 'Low' of Battery n1
' LedLow2   P2.5  A logic Hi light the Led 'Low' of Battery n2
' LedLow3   P2.6  A logic Hi light the Led 'Low' of Battery n3
' LedLow4   P2.7  A logic Hi light the Led 'Low' of Battery n4

' P3 Use ..................................................................

' In  DetBat1     P3.0  Activ Hi    A logic Hi indicate battery 1 present
' In  DetBat2     P3.1  Activ Hi    A logic Hi indicate battery 2 present
' In  DetBat3     P3.2  Activ Hi    A logic Hi indicate battery 3 present
' In  DetBat4     P3.3  Activ Hi    A logic Hi indicate battery 4 present
' In  CalMode     P3.4  Activ low   A logic Low indicate 'Calibration Mode'
' In  <Unused>    P3.5
' In  <Unused>    P3.6
' Out  Charge     P3.7  Activ low   A logic Low Switch on the charger

'------------------------------------------------------------------------------
$regfile = "89c51.dat"                              'Define machine
$crystal = 12000000                                 '12 MHz XTal


'Const declaration ------------------------------------------------------------

Const kDisAddNop = &H0                              'Adress of disp NonOp
Const kDisAddDecMode = &H9                          'Adress of disp DecodeMode
Const kDisAddLum = &HA                              'Adress of disp Intensity
Const kDisAddScanLim = &HB                          'Adress of disp ScanLim
Const kDisAddShutD = &HC                            'Adress of disp ShutD
Const kDisAddDisTest = &HF                          'Adress of disp Test

Const kDisSegMinus = &HA                            'Display Minus Sign
Const kDisSegBlank = &HF                            'Display Blank
Const kDisIntens = &HE                              'Seg Intensity

Const kBatScanPeriodDBug = 10                       'Period Debug in Sec
Const kBatChargePeriodDBug = 20                     'Period Debug in Sec

Const kBatScanPeriod = 300                          'Period Normal in Sec
Const kBatChargePeriod = 1800                       'Period Normal in Sec

Const kVoltLow = 126                                'Low voltage Limit
Const kVoltHi = 138                                 'Hight voltage Limit

'IO Aliasing ------------------------------------------------------------------

In_ADC Alias P0

Out_DispDIn Alias P1.0
Out_DispClock Alias P1.1
Out_DispLoad Alias P1.2

Out_AdcRead Alias P1.3
Out_AdcWrite Alias P1.4
In_AdcEOC Alias P1.5

In_DebugMode Alias P1.6
Out_WdTrig Alias P1.7

Out_BatLowPort Alias P2
Out_ActBat1 Alias P2.0
Out_ActBat2 Alias P2.1
Out_ActBat3 Alias P2.2
Out_ActBat4 Alias P2.3

Out_LedLow1 Alias P2.4
Out_LedLow2 Alias P2.5
Out_LedLow3 Alias P2.6
Out_LedLow4 Alias P2.7


In_DetBatQ Alias P3
In_DetBat1 Alias P3.0
In_DetBat2 Alias P3.1
In_DetBat3 Alias P3.2
In_DetBat4 Alias P3.3

In_CalMode Alias P3.4

Out_Charge Alias P3.7


'Var Decl ---------------------------------------------------------------------

Dim Clock_word As Word                              'Timer counter
Dim Tick_hsec As Bit                                'Every 500ms Tick flag
Dim Watchd As Bit                                   'Watchd toggle

Dim Volt As Byte                                    'Last DAC Voltage Read
Dim VoltEoc As Bit                                  'EOC State

Dim I As Byte , J As Byte , K As Byte               'Work Var
Dim BB As Bit                                       'Work Bit

Dim Bat As Byte                                     'Actual Bat Connected
Dim BV(4) As byte                                   'Bat State 0:No 1:Low 2:Ok

Dim PSel As byte                                    'Sel Quartet
Dim PLow As byte                                    'Low Quartet

Dim Period As Word                                  'Curent Period
Dim MaxPeriod As Word                               'Max Period

Dim CurBat As Byte                                  'Current Bat on charge
Dim NbBat As Byte                                   'Nbr Of Bat Low


Declare Sub DispVolt(DispVChan As Byte , DispVolt As Byte)
Dim DispVChan As Byte , DispVolt As Byte

Declare Sub WriteDisp(DispAdd As Byte , DispDat As Byte)
Dim DispAdd As Byte , DispDat As Byte

'Program ----------------------------------------------------------------------

OnReset:

'Configure Timer0 for Interupt every 250s

Config Timer0 = Timer , Gate = Internal , Mode = 2

On Timer0 Timer0_Int

Load Timer0 , 250

Priority Set Timer0
Enable Interrupts
Enable Timer0


'Init IO
Out_BatLowPort = &B00001111                         'Reset Low Port
Out_DispClock = 0                                   'Clear Disp Intf
Out_DispDIn = 0

'Init var
Clock_word = 0
Tick_hsec = 0
Watchd = 0

'INIT Timer
Start Timer0

'Init Display Ship

Call WriteDisp(kDisAddDecMode , &HFF)               'Use decoder for all digit
Call WriteDisp(kDisAddLum , kDisIntens)             'Set Up luminosity
Call WriteDisp(kDisAddScanLim , &H7)                'Use All Digit
Call WriteDisp(kDisAddShutD , &H1)                  'Wake it

'Test Display
Call WriteDisp(kDisAddDisTest , &H1)                'Display 188 to all

Waitmse 500                                         'Display Test for 500 ms

'End Test Display
Call WriteDisp(kDisAddDisTest , &H0)

' look for Cal Mode
BB = In_CalMode
If BB = 0 then
   Goto CalMode
EndIf

Bat = &HFF                                          'To insure init
CurBat = 4                                          'Start CurBat To 1


' Main Loop -------------------------------------------------------------------
Do
Main:
   If Tick_hsec = 1 Then                            'Every half Sec tick
      Tick_hsec = 0

      Out_WdTrig = 0                                'Pulse Watch dog
      Waitmse 2
      Out_WdTrig = 1

      ' Look if battery connected or disconected
      I = In_DetBatQ
      I = I And &B00001111
      J = &B00000001
      PSel = &B11111110
      PLow = 0

      If I <> Bat Then                              'Bat Cnx Changed
         Bat = I                                    'Update State
         NbBat = 0

         'Display minus on all Digit
         For I = 1 to 8
            Call WriteDisp(I , kDisSegMinus)
         Next I

         Out_Charge = 1                             'Stop Charge if any
         Waitmse 500                                'Give it time to sleep
         Out_WdTrig = 0                             'Toggle Watch Dog

         Out_BatLowPort = &B00001111                'Reset Low Port
         Waitmse 250                                'Wait for relay Establisment
         Out_WdTrig = 1                             'Toggle Watch Dog

         'Update BV and Display
         For I = 1 To 4
            K = Bat And J

            If K = 0 Then                           'If No Bat
               BV(I) = 0                            'Flag It To 0
            Else
               K = PSel AND &B00001111
               Out_BatLowPort = K                   'Sel Bat

               Out_WdTrig = 0                       'Toggle Watch Dog
               Waitmse 250                          'Wait for relay Establisment

               Gosub ReadAdc                        'Read Voltage
               Call DispVolt(I , Volt)              'Update Display
               BV(I) = 2                            'Update Array

               Out_BatLowPort = &B00001111          'No Sel, No Low Led for now
               Out_WdTrig = 1                       'Toggle Watch Dog
               Waitmse 250                          'Wait for relay Back

               If Volt < kVoltLow Then              'Set Low Flag if needed
                  Incr NbBat
                  BV(I) = 1
                  PLow = PLow Or J
               EndIf

            EndIf

            Shift J , Left , 1                      'Shift Mask For Next One
            Rotate PSel , Left , 1                  'Shift Sel For Next One
         Next I

         Shift PLow , Left , 4                      'Adjust PLow to Port

         BB = In_DebugMode                          'Look  if debug mode

         'Configure the item to charge if needed
         If NbBat <> 0 Then

            'Calc CharegPeriod
            If BB = 0 Then
               MaxPeriod = kBatChargePeriodDBug * 2
            Else
               MaxPeriod = kBatChargePeriod * 2
            EndIf

             'Calc CurBat and select it if needed
            For I = 1 To 4
               CurBat = CurBat + 1
               If CurBat > 4 Then
                  CurBat = 1
               EndIf

               If BV(CurBat) = 1 Then               'Bat Found

                  PSel = &B01111111
                  Rotate PSel , Left , CurBat       'Sel CurBat
                  J = PSel And &B00001111
                  J = J Or PLow                     'Display Low
                  Out_BatLowPort = J
                  Waitmse 400                       'Wait for relay

                  Out_Charge = 0                    'Charge It

                  Exit For
               EndIf
            Next I


         Else
            'Calc Scan Period
            If BB = 0 then
               MaxPeriod = kBatScanPeriodDBug * 2
            Else
               MaxPeriod = kBatScanPeriod * 2
            EndIf
         EndIf

         Period = 1                                 'Insure Start a new Cycle

         Goto Main                                  'So now loop normally
      EndIf


     'Look if Period End
      Period = Period + 1
      If Period > MaxPeriod Then                    'If Period ends, Scan Again
         Bat = &HFF
         Goto Main
      EndIf

      If NbBat = 0 Then                             'If not charge
          Goto Main                                 'Nothing To DO
      EndIf

      Gosub ReadAdc                                 'Read Voltage under charge
      Call DispVolt(CurBat , Volt)                  'Update Display

      If Volt >= kVoltHi Then                       'Avoid over load
         Period = MaxPeriod
      EndIf

   EndIf                                            'Of Tick


'And continue for ever, that's the computer hard duty..
Loop


'==============================================================================
' Calibration mode ============================================================
'==============================================================================
CalMode:

'Blank All display
For I = 1 to 8
   Call WriteDisp(I , kDisSegBlank)
Next I


Do

   If Tick_hsec = 1 Then                            'Every half Sec tick
      Tick_hsec = 0
      If Watchd = 0 Then
         Out_WdTrig = 1                             'Toggle Watch dog
         Watchd = 1
      Else
          Out_WdTrig = 0
          Watchd = 0
       EndIf

    'Read Voltage
    Gosub ReadAdc

    'Display Voltage
    Call DispVolt(1 , Volt)

   EndIf                                            'Of Tick


'And continue for ever, that's the computer hard duty..
Loop

'------------------------------------------------------------------------------
' Read Voltage at ADC and Fill Volt with the Value  ---------------------------
'------------------------------------------------------------------------------

ReadAdc:

   'Pulse Write to Init Measure
   Out_AdcWrite = 0
   Out_AdcWrite = 1

   ' Wait for measure Ok
   AdcWait:
   VoltEoc = In_AdcEOC
   If VoltEoc = 1 Then : Goto AdcWait : EndIf

   ' Read ADC Value
   Out_AdcRead = 0
   Volt = In_ADC
   Out_AdcRead = 1

Return

'------------------------------------------------------------------------------
' Timer 0 Interrupt -----------------------------------------------------------
'------------------------------------------------------------------------------

Timer0_Int:

'Incr Sec Ticks
Incr Clock_word
If Clock_word = 2000 Then
   Clock_word = 0
   Tick_hsec = 1
End If

Return

'------------------------------------------------------------------------------
' Display Voltage DispVolt in Chan DispVChan ----------------------------------
'------------------------------------------------------------------------------

Sub DispVolt(DispVChan As Byte , DispVolt As Byte)

Dim _DisV As Byte

'Get Digit Lo
DispVChan = DispVChan * 2

_DisV = DispVolt / 10
If _DisV > 9 then
    _DisV = _DisV - 10
    _DisV = _DisV OR &B10000000
EndIf
Call WriteDisp(DispVChan , _DisV)

'Disp volt
DispVChan = DispVChan - 1

'Disp 1/10 volt
_DisV = DispVolt Mod 10
Call WriteDisp(DispVChan , _DisV)

End Sub


'------------------------------------------------------------------------------
' Write in the digit controleur Adress, data  ---------------------------------
'------------------------------------------------------------------------------

Sub WriteDisp(DispAdd As Byte , DispDat As Byte)

Dim _DisW As Word
Dim _DigD as Word
Dim _Dig As Byte

'Build the serial data stream Adress + Data
_DisW = DispAdd * 256
_DisW = _DisW + DispDat

'Set Load Low to start the data Burst
Out_DispLoad = 0

'Serialize the data
For _Dig = 1 To 16
   _DigD = _DisW And &B1000000000000000
   If _DigD = &B1000000000000000 Then
      Out_DispDIn = 1
   Else
      Out_DispDIn = 0
   End If

   'Pulse Clock
   Out_DispClock = 1
   Out_DispClock = 0


   'Treat next bit
   Shift _DisW , Left , 1
Next _i

'Restore Din to low State
Out_DispDIn = 0


'Set Load Hi to End the data Burst and load Data to Display register
Out_DispLoad = 1

End Sub